import { GraphQLError, Kind, parse, print, isTypeDefinitionNode, visit, } from 'graphql';
import { getAutocompleteSuggestions, getHoverInformation, validateQuery, getRange, DIAGNOSTIC_SEVERITY, getOutline, getDefinitionQueryResultForFragmentSpread, getDefinitionQueryResultForDefinitionNode, getDefinitionQueryResultForNamedType, getDefinitionQueryResultForField, getASTNodeAtPosition, getTokenAtPosition, getTypeInfo, Range, Position, } from 'graphql-language-service';
import { SymbolKind, } from 'vscode-languageserver-types';
const KIND_TO_SYMBOL_KIND = {
    [Kind.FIELD]: SymbolKind.Field,
    [Kind.OPERATION_DEFINITION]: SymbolKind.Class,
    [Kind.FRAGMENT_DEFINITION]: SymbolKind.Class,
    [Kind.FRAGMENT_SPREAD]: SymbolKind.Struct,
    [Kind.OBJECT_TYPE_DEFINITION]: SymbolKind.Class,
    [Kind.ENUM_TYPE_DEFINITION]: SymbolKind.Enum,
    [Kind.ENUM_VALUE_DEFINITION]: SymbolKind.EnumMember,
    [Kind.INPUT_OBJECT_TYPE_DEFINITION]: SymbolKind.Class,
    [Kind.INPUT_VALUE_DEFINITION]: SymbolKind.Field,
    [Kind.FIELD_DEFINITION]: SymbolKind.Field,
    [Kind.INTERFACE_TYPE_DEFINITION]: SymbolKind.Interface,
    [Kind.DOCUMENT]: SymbolKind.File,
    FieldWithArguments: SymbolKind.Method,
};
function getKind(tree) {
    if (tree.kind === 'FieldDefinition' &&
        tree.children &&
        tree.children.length > 0) {
        return KIND_TO_SYMBOL_KIND.FieldWithArguments;
    }
    return KIND_TO_SYMBOL_KIND[tree.kind];
}
export class GraphQLLanguageService {
    constructor(cache, logger) {
        this._graphQLCache = cache;
        this._graphQLConfig = cache.getGraphQLConfig();
        this._logger = logger;
    }
    getConfigForURI(uri) {
        const config = this._graphQLCache.getProjectForFile(uri);
        if (config) {
            return config;
        }
    }
    async getDiagnostics(document, uri, isRelayCompatMode) {
        var _a, _b;
        let documentHasExtensions = false;
        const projectConfig = this.getConfigForURI(uri);
        if (!projectConfig || !document || document.trim().length < 2) {
            return [];
        }
        const { schema: schemaPath, name: projectName, extensions } = projectConfig;
        try {
            const documentAST = parse(document);
            if (!schemaPath || uri !== schemaPath) {
                documentHasExtensions = documentAST.definitions.some(definition => {
                    switch (definition.kind) {
                        case Kind.OBJECT_TYPE_DEFINITION:
                        case Kind.INTERFACE_TYPE_DEFINITION:
                        case Kind.ENUM_TYPE_DEFINITION:
                        case Kind.UNION_TYPE_DEFINITION:
                        case Kind.SCALAR_TYPE_DEFINITION:
                        case Kind.INPUT_OBJECT_TYPE_DEFINITION:
                        case Kind.SCALAR_TYPE_EXTENSION:
                        case Kind.OBJECT_TYPE_EXTENSION:
                        case Kind.INTERFACE_TYPE_EXTENSION:
                        case Kind.UNION_TYPE_EXTENSION:
                        case Kind.ENUM_TYPE_EXTENSION:
                        case Kind.INPUT_OBJECT_TYPE_EXTENSION:
                        case Kind.DIRECTIVE_DEFINITION:
                            return true;
                    }
                    return false;
                });
            }
        }
        catch (error) {
            if (error instanceof GraphQLError) {
                const range = getRange((_b = (_a = error.locations) === null || _a === void 0 ? void 0 : _a[0]) !== null && _b !== void 0 ? _b : { column: 0, line: 0 }, document);
                return [
                    {
                        severity: DIAGNOSTIC_SEVERITY.Error,
                        message: error.message,
                        source: 'GraphQL: Syntax',
                        range,
                    },
                ];
            }
            throw error;
        }
        let source = document;
        const fragmentDefinitions = await this._graphQLCache.getFragmentDefinitions(projectConfig);
        const fragmentDependencies = await this._graphQLCache.getFragmentDependencies(document, fragmentDefinitions);
        const dependenciesSource = fragmentDependencies.reduce((prev, cur) => `${prev} ${print(cur.definition)}`, '');
        source = `${source} ${dependenciesSource}`;
        let validationAst = null;
        try {
            validationAst = parse(source);
        }
        catch (_c) {
            return [];
        }
        let customRules = null;
        if ((extensions === null || extensions === void 0 ? void 0 : extensions.customValidationRules) &&
            typeof extensions.customValidationRules === 'function') {
            customRules = extensions.customValidationRules(this._graphQLConfig);
        }
        const schema = await this._graphQLCache.getSchema(projectName, documentHasExtensions);
        if (!schema) {
            return [];
        }
        return validateQuery(validationAst, schema, customRules, isRelayCompatMode);
    }
    async getAutocompleteSuggestions(query, position, filePath) {
        var _a, _b, _c;
        const projectConfig = this.getConfigForURI(filePath);
        if (!projectConfig) {
            return [];
        }
        const schema = await this._graphQLCache.getSchema(projectConfig.name);
        const fragmentDefinitions = await this._graphQLCache.getFragmentDefinitions(projectConfig);
        const fragmentInfo = Array.from(fragmentDefinitions).map(([, info]) => info.definition);
        if (schema) {
            return getAutocompleteSuggestions(schema, query, position, undefined, fragmentInfo, {
                uri: filePath,
                fillLeafsOnComplete: (_c = (_b = (_a = projectConfig === null || projectConfig === void 0 ? void 0 : projectConfig.extensions) === null || _a === void 0 ? void 0 : _a.languageService) === null || _b === void 0 ? void 0 : _b.fillLeafsOnComplete) !== null && _c !== void 0 ? _c : false,
            });
        }
        return [];
    }
    async getHoverInformation(document, position, filePath, options) {
        const projectConfig = this.getConfigForURI(filePath);
        if (!projectConfig) {
            return '';
        }
        const schema = await this._graphQLCache.getSchema(projectConfig.name);
        if (schema) {
            return getHoverInformation(schema, document, position, undefined, options);
        }
        return '';
    }
    async getDefinition(document, position, filePath) {
        const projectConfig = this.getConfigForURI(filePath);
        if (!projectConfig) {
            return null;
        }
        let ast;
        try {
            ast = parse(document);
        }
        catch (_a) {
            return null;
        }
        const node = getASTNodeAtPosition(document, ast, position);
        if (node) {
            switch (node.kind) {
                case Kind.FRAGMENT_SPREAD:
                    return this._getDefinitionForFragmentSpread(document, ast, node, filePath, projectConfig);
                case Kind.FRAGMENT_DEFINITION:
                case Kind.OPERATION_DEFINITION:
                    return getDefinitionQueryResultForDefinitionNode(filePath, document, node);
                case Kind.NAMED_TYPE:
                    return this._getDefinitionForNamedType(document, ast, node, filePath, projectConfig);
                case Kind.FIELD:
                    return this._getDefinitionForField(document, ast, node, filePath, projectConfig, position);
            }
        }
        return null;
    }
    async getDocumentSymbols(document, filePath) {
        const outline = await this.getOutline(document);
        if (!outline) {
            return [];
        }
        const output = [];
        const input = outline.outlineTrees.map((tree) => [null, tree]);
        while (input.length > 0) {
            const res = input.pop();
            if (!res) {
                return [];
            }
            const [parent, tree] = res;
            if (!tree) {
                return [];
            }
            output.push({
                name: tree.representativeName,
                kind: getKind(tree),
                location: {
                    uri: filePath,
                    range: {
                        start: tree.startPosition,
                        end: tree.endPosition,
                    },
                },
                containerName: parent ? parent.representativeName : undefined,
            });
            input.push(...tree.children.map(child => [tree, child]));
        }
        return output;
    }
    async getReferences(document, position, filePath) {
        var _a;
        const projectConfig = this.getConfigForURI(filePath);
        if (!projectConfig) {
            return [];
        }
        let ast;
        try {
            ast = parse(document);
        }
        catch (_b) {
            return [];
        }
        const definitionNode = getASTNodeAtPosition(document, ast, position);
        const name = (_a = definitionNode === null || definitionNode === void 0 ? void 0 : definitionNode.name) === null || _a === void 0 ? void 0 : _a.value;
        if (!name) {
            return [];
        }
        const schema = await projectConfig.getSchema('DocumentNode');
        if (!schema) {
            return [];
        }
        const references = [];
        const matchNodeByName = (node) => {
            var _a;
            if (!((_a = node.loc) === null || _a === void 0 ? void 0 : _a.source)) {
                return;
            }
            if (node.name.value.toLowerCase() === name.toLowerCase()) {
                references.push({
                    location: {
                        uri: node.loc.source.name,
                        range: new Range(new Position(node.loc.startToken.line - 1, node.loc.startToken.column - 1), new Position(node.loc.endToken.line - 1, node.loc.endToken.column - 1)),
                    },
                });
            }
        };
        visit(parse(document), {
            FragmentSpread(node) {
                matchNodeByName(node);
            },
            NamedType(node) {
                matchNodeByName(node);
            },
        });
        visit(schema, {
            FragmentSpread(node) {
                matchNodeByName(node);
            },
            NamedType(node) {
                matchNodeByName(node);
            },
        });
        return references;
    }
    async _getDefinitionForNamedType(document, ast, node, filePath, projectConfig) {
        const objectTypeDefinitions = await this._graphQLCache.getObjectTypeDefinitions(projectConfig);
        const dependencies = await this._graphQLCache.getObjectTypeDependenciesForAST(ast, objectTypeDefinitions);
        const localOperationDefinitionInfos = ast.definitions
            .filter(isTypeDefinitionNode)
            .map((definition) => ({
            filePath,
            content: document,
            definition,
        }));
        const result = await getDefinitionQueryResultForNamedType(document, node, dependencies.concat(localOperationDefinitionInfos));
        return result;
    }
    async _getDefinitionForField(document, _ast, _node, _filePath, projectConfig, position) {
        var _a;
        const token = getTokenAtPosition(document, position);
        const schema = await this._graphQLCache.getSchema(projectConfig.name);
        const typeInfo = getTypeInfo(schema, token.state);
        const fieldName = (_a = typeInfo.fieldDef) === null || _a === void 0 ? void 0 : _a.name;
        if (typeInfo && fieldName) {
            const parentTypeName = typeInfo.parentType.toString();
            const objectTypeDefinitions = await this._graphQLCache.getObjectTypeDefinitions(projectConfig);
            const dependencies = [...objectTypeDefinitions.values()];
            const result = await getDefinitionQueryResultForField(fieldName, parentTypeName, dependencies);
            return result;
        }
        return null;
    }
    async _getDefinitionForFragmentSpread(document, ast, node, filePath, projectConfig) {
        const fragmentDefinitions = await this._graphQLCache.getFragmentDefinitions(projectConfig);
        const dependencies = await this._graphQLCache.getFragmentDependenciesForAST(ast, fragmentDefinitions);
        const localFragDefinitions = ast.definitions.filter(definition => definition.kind === Kind.FRAGMENT_DEFINITION);
        const typeCastedDefs = localFragDefinitions;
        const localFragInfos = typeCastedDefs.map((definition) => ({
            filePath,
            content: document,
            definition,
        }));
        const result = await getDefinitionQueryResultForFragmentSpread(document, node, dependencies.concat(localFragInfos));
        return result;
    }
    async getOutline(document) {
        return getOutline(document);
    }
}
//# sourceMappingURL=GraphQLLanguageService.js.map